from functools import partial, wraps

from .constants import ValidLogLevels
from .functions import (
    _disable_logging,
    _is_valid_log_level_int,
    _reset_log_emergence_level,
    log_level,
)

# Identifier for '_inplace_enabled' to use as a (temporary) attribute name
INPLACE_ENABLED_PLACEHOLDER = "_inplace_store"


def _inplace_enabled(operation_method=None, *, default=False):
    """A decorator enabling operations to be applied in-place.

    If the decorated method has keyword argument *inplace* being equal
    to True, the function will be performed on `self` and return None,
    otherwise it will operate on a copy of ``self`` and return the
    processed copy.

    Note that methods decorated with this should assign the core
    variable storing the relevant instance for use throughout the
    method to ``_inplace_enabled_define_and_cleanup(self)``.

    :Parameters:

        operation_method: method

        default: `bool`

    """

    def decorator(operation_method, default=False):
        @wraps(operation_method)
        def inplace_wrapper(self, *args, **kwargs):
            is_inplace = kwargs.get("inplace", default)
            if is_inplace:
                self.INPLACE_ENABLED_PLACEHOLDER = self
            else:
                self.INPLACE_ENABLED_PLACEHOLDER = self.copy()

            processed_copy = operation_method(self, *args, **kwargs)

            if is_inplace:
                return  # decorated function returns None in this case
            else:
                return processed_copy

        return inplace_wrapper

    if operation_method is None:
        return partial(decorator, default=default)

    return decorator


def _inplace_enabled_define_and_cleanup(instance):
    """Deletes attribute set by _inplace_enabled, returning its value.

    Designed as a convenience function for use at the start of methods
    decorated by _inplace_enabled; the core variable used throughout
    for the instance in the decorated method should first be assigned
    to this function with the class instance as the input. For
    example:

    d = _inplace_enabled_define_and_cleanup(self)

    should be set initially for a method operating inplace or
    otherwise (via inplace_enabled) on a data array, d.

    In doing so, the relevant construct variable can be defined
    appropriately and the internal attribute created for that purpose
    by inplace_enabled (which is no longer required) can be cleaned
    up, all in one line.

    """
    try:
        x = instance._custom.pop(INPLACE_ENABLED_PLACEHOLDER)
    except (AttributeError, KeyError):
        x = instance.INPLACE_ENABLED_PLACEHOLDER
        del instance.INPLACE_ENABLED_PLACEHOLDER

    return x


def _manage_log_level_via_verbosity(method_with_verbose_kwarg, calls=[0]):
    """A decorator to manage log filtering by verbosity argument.

    This enables overriding of the log severity level such that an
    integer input (lying in the valid range) to the decorated function
    will ignore the global cfdm.log_level() to configure a custom
    verbosity for the individual function call, applying to its logic
    and any functions it calls internally and lasting only the
    duration of the call.

    If verbose=None, as is the default, the log_level() determines
    which log messages are shown, as standard.

    Only use this to decorate functions which make log calls directly
    and have a 'verbose' keyword argument set to None by default.

    Note that the 'calls' keyword argument is to automatically track
    the number of decorated functions that are being (or about to be)
    executed, with the purpose of preventing resetting of the
    effective log level at the completion of decorated functions that
    are called inside other decorated functions (see comments in
    'finally' statement for further explanation).  Note (when it is of
    concern) that this approach may not be thread-safe.

    """
    # Note that 'self' can be included in '*args' for any function calls
    # below, such that this decorator will work for both methods and
    # functions that are not bound to classes.

    @wraps(method_with_verbose_kwarg)
    def verbose_override_wrapper(*args, **kwargs):
        # Increment indicates that one decorated function has started
        # execution
        calls[0] += 1

        # Deliberately error if verbose kwarg not set, if not by user
        # then as a default to the decorated function, as this is
        # crucial to usage.
        verbose = kwargs.get("verbose")

        possible_levels = ", ".join(
            [val.name + " = " + str(val.value) for val in ValidLogLevels]
        )
        invalid_arg_msg = (
            f"Invalid value '{verbose}' for the 'verbose' keyword argument. "
            "Accepted values are integers corresponding in positive "
            f"cases to increasing verbosity (namely {possible_levels}), or "
            "None, to configure the verbosity according to the global "
            "log_level setting."
        )
        # First convert valid string inputs to the enum-mapped int constant:
        if isinstance(verbose, str):
            uppercase_arg = verbose.upper()
            if hasattr(ValidLogLevels, uppercase_arg):
                verbose = getattr(ValidLogLevels, uppercase_arg).value
            else:
                raise ValueError(invalid_arg_msg)

        # Convert Boolean cases for backwards compatibility. Need 'is'
        # identity rather than '==' (value) equivalency test, since 1
        # == True, etc.
        if verbose is True:
            verbose = 3  # max verbosity excluding debug levels
        elif verbose is False:
            verbose = 0  # corresponds to disabling logs i.e. no verbosity

        # Override log levels for the function & all it calls (to
        # reset at end)
        if verbose is not None:  # None as default, note exclude True & False
            if _is_valid_log_level_int(verbose):
                _reset_log_emergence_level(verbose)
            else:
                raise ValueError(invalid_arg_msg)

        # First need to (temporarily) re-enable global logging if
        # disabled in the cases where you do not want to disable it
        # anyway:
        if log_level() == "DISABLE" and verbose not in (0, None):
            _disable_logging(at_level="NOTSET")  # enables all logging again

        # After method completes, re-set any changes to log level or
        # enabling
        try:
            return method_with_verbose_kwarg(*args, **kwargs)
        except Exception:
            raise
        finally:  # so that crucial 'teardown' code runs even if
            # method errors Decrement indicates one decorated function
            # has finished execution
            calls[0] -= 1
            # Due to the incrementing and decrementing of 'calls', it
            # will only be zero when the outermost decorated method
            # has finished, so the following condition prevents
            # resetting occurring once inner functions complete (which
            # would mean any subsequent code in the outer function
            # would undesirably regain the global level):
            if calls[0] == 0:
                if verbose == 0:
                    _disable_logging(at_level="NOTSET")  # lift deactivation
                elif verbose is not None and _is_valid_log_level_int(verbose):
                    _reset_log_emergence_level(log_level())
                if log_level() == "DISABLE" and verbose != 0:
                    _disable_logging()  # disable again after re-enabling

    return verbose_override_wrapper


# @_test_decorator_args('i') -> example usage for decorating, using i kwarg
def _test_decorator_args(*dec_args):
    """A wrapper to provide positional arguments to the decorator."""

    def deprecated_kwarg_check_decorator(operation_method):
        """A decorator for a deprecation check on given kwargs.

        To specify deprecated kwargs, supply them as string arguments, e.g:

            @_test_decorator_args('i')
            @_test_decorator_args('i', 'traceback')

        For a specified list `*dec_args`, check if the decorated
        method has been supplied with any of the elements as keyword
        arguments and if so, call _DEPRECATION_ERROR_KWARGS on them,
        optionally providing a custom message to raise inside it.

        """

        @wraps(operation_method)
        def precede_with_kwarg_deprecation_check(self, *args, **kwargs):
            print(
                "In precede_with_kwarg_deprecation_check. dec_args=", dec_args
            )

            # Decorated method has same return signature as if undecorated:
            return

        return precede_with_kwarg_deprecation_check

    return deprecated_kwarg_check_decorator


def _display_or_return(method_with_display_kwarg):
    """A decorator enabling a string to be printed rather than returned.

    If the decorated method has keyword argument *display* being equal
    to True, by default or from being set as such, the function will
    print the output that would otherwise be returned and return `None`.

    :Parameters:

        method_with_display_kwarg: method

    """

    @wraps(method_with_display_kwarg)
    def end_with_display_or_return_logic(self, *args, **kwargs):
        string = method_with_display_kwarg(self, *args, **kwargs)

        # display=True is always default, so if display not provided, set True
        display = kwargs.get("display", True)

        if display:
            print(string)
        else:
            return string

    return end_with_display_or_return_logic



from .propertiesdata import PropertiesData


class Topology(PropertiesData):
    """Abstract base class for topology-related constructs.

    .. versionadded:: (cfdm) 1.11.0.0

    """

    @property
    def ndim(self):
        """The number of data dimensions.

        Only the data dimensions that correspond to a domain axis
        construct are included.

        .. versionadded:: (cfdm) 1.11.0.0

        .. seealso:: `data`, `has_data`, `shape`, `size`

        **Examples**

        >>> d.shape
        (1324,)
        >>> d.ndim
        1
        >>> f.size
        1324

        """
        try:
            return len(self.shape)
        except AttributeError:
            raise AttributeError(
                f"{self.__class__.__name__} object has no attribute 'ndim'"
            )

    @property
    def shape(self):
        """A tuple of the data array's dimension sizes.

        Only the data dimension that corresponds to a domain axis
        construct is included.

        .. versionadded:: (cfdm) 1.11.0.0

        .. seealso:: `data`, `has_data`, `ndim`, `size`

        **Examples**

        >>> d.shape
        (1324,)
        >>> d.ndim
        1
        >>> d.size
        1324

        """
        data = self.get_data(None, _units=False, _fill_value=False)
        if data is not None:
            return data.shape[:1]

        raise AttributeError(
            f"{self.__class__.__name__} object has no attribute 'shape'"
        )

    @property
    def size(self):
        """The number elements in the data.

        `size` is equal to the product of `shape`, that only includes
        the data dimension corresponding to a domain axis construct.

        .. versionadded:: (cfdm) 1.11.0.0

        .. seealso:: `data`, `has_data`, `ndim`, `shape`

        **Examples**

        >>> d.shape
        (1324,)
        >>> d.ndim
        1
        >>> d.size
        1324

        """
        try:
            return self.shape[0]
        except AttributeError:
            raise AttributeError(
                f"{self.__class__.__name__} object has no attribute 'size'"
            )



import copy
import datetime
import unittest

import numpy as np

import cfdm


class Core2Test(unittest.TestCase):
    """Unit test for the `cfdm.core` package."""

    def setUp(self):
        """Preparations called immediately before each test method."""
        # Disable log messages to silence expected warnings
        cfdm.log_level("DISABLE")
        # Note: to enable all messages for given methods, lines or
        # calls (those without a 'verbose' option to do the same)
        # e.g. to debug them, wrap them (for methods, start-to-end
        # internally) as follows:
        #
        # cfdm.log_level('DEBUG')
        # < ... test code ... >
        # cfdm.log_level('DISABLE')

    def test_core_NumpyArray(self):
        """Test cfdm.core.NumpyArray class."""
        a = cfdm.core.NumpyArray(np.array([1, 2, 3]))

        # __deepcopy__
        copy.deepcopy(a)

        # __array__
        np.array(a)
        np.array(a, dtype="float")

    def test_core_Container(self):
        """Test cfdm.core.Container class."""
        cfdm.core.Container(source="qwerty")


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cfdm.environment()
    print("")
    unittest.main(verbosity=2)



import numpy as np

from ..functions import indices_shape, parse_indices
from .abstract import Array
from .mixin import IndexMixin
from .mixin.arraymixin import array_implements


class FullArray(IndexMixin, Array):
    """A array filled with a given value.

    The array may be empty or all missing values.

    .. versionadded:: (cfdm) 1.12.0.0

    """

    # Copy the functions handled by __array_function__ implementations
    # (numpy NEP 18) from Array, so that they can be extended.
    _HANDLED_FUNCTIONS = Array._HANDLED_FUNCTIONS.copy()

    def __init__(
        self,
        fill_value=None,
        dtype=None,
        shape=None,
        attributes=None,
        source=None,
        copy=True,
    ):
        """**Initialisation**

        :Parameters:

            fill_value : scalar, optional
                The fill value for the array. May be set to
                `cf.masked` or `np.ma.masked`.

            dtype: `numpy.dtype`
                The data type of the array.

            shape: `tuple`
                The array dimension sizes.

            {{init attributes: `dict` or `None`, optional}}

            {{init source: optional}}

            {{init copy: `bool`, optional}}

        """
        super().__init__(source=source, copy=copy)

        if source is not None:
            try:
                fill_value = source._get_component("full_value", None)
            except AttributeError:
                fill_value = None

            try:
                dtype = source._get_component("dtype", None)
            except AttributeError:
                dtype = None

            try:
                shape = source._get_component("shape", None)
            except AttributeError:
                shape = None

            try:
                attributes = source._get_component("attributes", False)
            except AttributeError:
                attributes = None

        self._set_component("full_value", fill_value, copy=False)
        self._set_component("dtype", dtype, copy=False)
        self._set_component("shape", shape, copy=False)
        self._set_component("attributes", attributes, copy=False)

    def __repr__(self):
        """Called by the `repr` built-in function.

        x.__repr__() <==> repr(x)

        """
        return f"<CF {self.__class__.__name__}{self.shape}: {self}>"

    def __str__(self):
        """Called by the `str` built-in function.

        x.__str__() <==> str(x)

        """
        fill_value = self.get_full_value()
        if fill_value is None:
            return "Uninitialised"

        return f"Filled with {fill_value!r}"

    def _get_array(self, index=None):
        """Returns the data as a `numpy` array.

        .. versionadded:: (cfdm) 1.12.0.0

        .. seealso:: `__array__`, `index`

        :Parameters:

            {{index: `tuple` or `None`, optional}}

        :Returns:

            `numpy.ndarray`
                The data.

        """
        if index is None:
            shape = self.shape
        else:
            original_shape = self.original_shape
            index = parse_indices(original_shape, index, keepdims=False)
            shape = indices_shape(index, original_shape, keepdims=False)

        fill_value = self.get_full_value()
        if fill_value is np.ma.masked:
            array = np.ma.masked_all(shape, dtype=self.dtype)
        elif fill_value is not None:
            array = np.full(shape, fill_value=fill_value, dtype=self.dtype)
        else:
            array = np.empty(shape, dtype=self.dtype)

        return array

    @property
    def array(self):
        """Return an independent numpy array containing the data.

        .. versionadded:: (cfdm) 1.12.0.0

        :Returns:

            `numpy.ndarray`
                An independent numpy array of the data.

        """
        return self._get_array()

    @property
    def dtype(self):
        """Data-type of the data elements."""
        return self._get_component("dtype")

    @property
    def shape(self):
        """Tuple of array dimension sizes."""
        return self._get_component("shape")

    def get_full_value(self, default=AttributeError()):
        """Return the data array fill value.

        .. versionadded:: (cfdm) 1.12.0.0

        .. seealso:: `set_full_value`

        :Parameters:

            default: optional
                Return the value of the *default* parameter if the
                fill value has not been set. If set to an `Exception`
                instance then it will be raised instead.

        :Returns:

                The fill value.

        """
        return self._get_component("full_value", default=default)

    def set_full_value(self, fill_value):
        """Set the data array fill value.

        .. versionadded:: (cfdm) 1.12.0.0

        .. seealso:: `get_full_value`

        :Parameters:

            fill_value : scalar, optional
                The fill value for the array. May be set to
                `cf.masked` or `np.ma.masked`.

        :Returns:

            `None`

        """
        self._set_component("full_value", fill_value, copy=False)


# --------------------------------------------------------------------
# __array_function__ implementations (numpy NEP 18)
# --------------------------------------------------------------------
@array_implements(FullArray, np.unique)
def unique(
    a, return_index=False, return_inverse=False, return_counts=False, axis=None
):
    """Version of `np.unique` that is optimised for `FullArray` objects.

    .. versionadded:: (cfdm) 1.12.0.0

    """
    if return_index or return_inverse or return_counts or axis is not None:
        # Fall back to the slow unique. (I'm sure we could probably do
        # something more clever here, but there is no use case at
        # present.)
        return np.unique(
            a[...],
            return_index=return_index,
            return_inverse=return_inverse,
            return_counts=return_counts,
            axis=axis,
        )

    # Efficient unique based on the full value
    x = a.get_full_value()
    if x is np.ma.masked:
        return np.ma.masked_all((1,), dtype=a.dtype)

    return np.array((x,), dtype=a.dtype)



import datetime
import faulthandler
import unittest

faulthandler.enable()  # to debug seg faults and timeouts

import cfdm


class DomainAxisTest(unittest.TestCase):
    """Unit test for the DomainAxis class."""

    d = cfdm.DomainAxis(size=99)
    d.nc_set_dimension("ncdim")

    def setUp(self):
        """Preparations called immediately before each test method."""
        # Disable log messages to silence expected warnings
        cfdm.log_level("DISABLE")
        # Note: to enable all messages for given methods, lines or
        # calls (those without a 'verbose' option to do the same)
        # e.g. to debug them, wrap them (for methods, start-to-end
        # internally) as follows:
        #
        # cfdm.LOG_LEVEL('DEBUG')
        # < ... test code ... >
        # cfdm.log_level('DISABLE')

    def test_DomainAxis__repr__str_construct_type(self):
        """Test DomainAxis inspection and `construct_type` method."""
        d = self.d

        repr(d)
        str(d)
        self.assertEqual(d.construct_type, "domain_axis")

    def test_DomainAxis_source(self):
        """Test the source keyword argument to DomainAxis."""
        d = self.d

        self.assertTrue(d.equals(cfdm.DomainAxis(source=d)))

        self.assertIsInstance(
            cfdm.DomainAxis(source="QWERTY"), cfdm.DomainAxis
        )

    def test_DomainAxis_equals(self):
        """Test the equality-testing DomainAxis method."""
        d = self.d
        e = d.copy()

        self.assertTrue(d.equals(d, verbose=3))
        self.assertTrue(d.equals(e, verbose=3))
        self.assertTrue(e.equals(d, verbose=3))

    def test_DomainAxis_size(self):
        """Test the size access and (un)setting DomainAxis method."""
        d = self.d.copy()
        d.set_size(100)

        self.assertTrue(d.has_size())
        self.assertEqual(d.get_size(), 100)
        self.assertEqual(d.del_size(), 100)
        self.assertIsNone(d.get_size(None))
        self.assertIsNone(d.del_size(None))
        self.assertFalse(d.has_size())

    def test_DomainAxis_unlimited(self):
        """Test the netCDF unlimited DomainAxis methods."""
        d = cfdm.DomainAxis()
        d.set_size(99)

        self.assertFalse(d.nc_is_unlimited())

        d.nc_set_unlimited(False)
        self.assertFalse(d.nc_is_unlimited())

        d.nc_set_unlimited(True)
        self.assertTrue(d.nc_is_unlimited())

        d.nc_set_unlimited(False)
        self.assertFalse(d.nc_is_unlimited())


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cfdm.environment()
    print()
    unittest.main(verbosity=2)



import abc


class IO(metaclass=abc.ABCMeta):
    """Abstract base class for reading and writing Fields."""

    def __init__(self, implementation):
        """**Initialisation**

        :Parameters:

            implementation: `Implementation'
                The objects required to represent a Field.

        """
        self.implementation = implementation

    @abc.abstractmethod
    def file_close(self, *args, **kwargs):
        """Close the dataset file."""
        raise NotImplementedError()  # pragma: no cover

    @abc.abstractmethod
    def file_open(self, *args, **kwargs):
        """Open the dataset file."""
        raise NotImplementedError()  # pragma: no cover


class IORead(IO, metaclass=abc.ABCMeta):
    """Abstract base class for instantiating Fields from a dataset."""

    @abc.abstractmethod
    def read(self, *args, **kwargs):
        """Read fields from a netCDF file."""
        raise NotImplementedError()  # pragma: no cover


class IOWrite(IO, metaclass=abc.ABCMeta):
    """Abstract base class for writing Fields to a dataset."""

    @abc.abstractmethod
    def write(self, *args, **kwargs):
        """Write fields to a netCDF file."""
        raise NotImplementedError()  # pragma: no cover



from . import core, mixin


class InterpolationParameter(
    mixin.NetCDFVariable,
    mixin.PropertiesData,
    mixin.Files,
    core.abstract.PropertiesData,
):
    """An interpolation parameter variable.

    Space may be saved by storing a subsample of the coordinates. The
    uncompressed coordinates can be reconstituted by interpolation
    from the subsampled coordinate values, also called either "tie
    points" or "bounds tie points".

    An interpolation parameter variable provides values for
    coefficient terms in the interpolation equation, or for any other
    terms that configure the interpolation process.

    **NetCDF interface**

    {{netCDF variable}}

    {{netCDF HDF5 chunks}}

    The netCDF subsampled dimension name and the netCDF interpolation
    subarea dimension name, if required, are set on the on the
    corresponding tie point index variable.

    .. versionadded:: (cfdm) 1.10.0.0

    .. seealso:: `TiePointIndex`

    """

    def __init__(
        self,
        properties=None,
        data=None,
        source=None,
        copy=True,
        _use_data=True,
    ):
        """**Initialisation**

        :Parameters:

            {{init properties: `dict`, optional}}

                *Parameter example:*
                  ``properties={'long_name': 'interpolation parameter'}``

            {{init data: data_like, optional}}

            {{init source: optional}}

            {{init copy: `bool`, optional}}

        """
        super().__init__(
            properties=properties,
            data=data,
            source=source,
            copy=copy,
            _use_data=_use_data,
        )

        self._initialise_netcdf(source)
        self._initialise_original_filenames(source)

    def dump(
        self,
        display=True,
        _key=None,
        _title=None,
        _create_title=True,
        _prefix="",
        _level=0,
        _omit_properties=None,
        _axes=None,
        _axis_names=None,
    ):
        """A full description of the interpolation parameter variable.

        Returns a description of all properties, including those of
        components, and provides selected values of all data arrays.

        .. versionadded:: (cfdm) 1.10.0.0

        :Parameters:

            display: `bool`, optional
                If False then return the description as a string. By
                default the description is printed.

        :Returns:

            {{returns dump}}

        """
        if _create_title and _title is None:
            _title = "Interpolation parameter: " + self.identity(default="")

        return super().dump(
            display=display,
            _key=_key,
            _omit_properties=_omit_properties,
            _prefix=_prefix,
            _level=_level,
            _title=_title,
            _create_title=_create_title,
            _axes=_axes,
            _axis_names=_axis_names,
        )



import numpy as np

from .abstract import Subarray


class RaggedSubarray(Subarray):
    """A subarray of a compressed ragged array.

    A subarray describes a unique part of the uncompressed array.

    See CF section 9 "Discrete Sampling Geometries".

    .. versionadded:: (cfdm) 1.10.0.0

    """

    def __getitem__(self, indices):
        """Return a subspace of the uncompressed subarray.

        x.__getitem__(indices) <==> x[indices]

        Returns a subspace of the uncompressed subarray as an
        independent numpy array.

        .. versionadded:: (cfdm) 1.10.0.0

        """
        d1, u_dims = self.compressed_dimensions().popitem()
        uncompressed_shape = self.shape

        data = self._select_data(check_mask=False)

        if data.size:
            shape = list(data.shape)
            u_indices0 = [slice(None)] * data.ndim

            u_indices0[d1] = slice(0, shape[d1])
            shape[d1] = uncompressed_shape[u_dims[-1]]

            u = np.ma.masked_all(shape, dtype=self.dtype)

            u[tuple(u_indices0)] = data

            u = u.reshape(uncompressed_shape)
        else:
            # This subarray contains no elements of the compressed
            # data
            u = np.ma.masked_all(uncompressed_shape, dtype=self.dtype)

        if indices is Ellipsis:
            return u

        return u[indices]

    @property
    def dtype(self):
        """The data-type of the uncompressed data.

        .. versionadded:: (cfdm) 1.10.0.0

        """
        return self.data.dtype



from .abstract import Array, CompressedArray, MeshArray, RaggedArray

from .subarray import (
    BiLinearSubarray,
    BiQuadraticLatitudeLongitudeSubarray,
    BoundsFromNodesSubarray,
    CellConnectivitySubarray,
    GatheredSubarray,
    InterpolationSubarray,
    LinearSubarray,
    QuadraticLatitudeLongitudeSubarray,
    QuadraticSubarray,
    RaggedSubarray,
)

from .subarray.abstract import MeshSubarray, Subarray, SubsampledSubarray

from .aggregatedarray import AggregatedArray
from .boundsfromnodesarray import BoundsFromNodesArray
from .cellconnectivityarray import CellConnectivityArray
from .gatheredarray import GatheredArray
from .fullarray import FullArray
from .h5netcdfarray import H5netcdfArray
from .netcdfarray import NetCDFArray
from .netcdf4array import NetCDF4Array
from .netcdfindexer import netcdf_indexer
from .numpyarray import NumpyArray
from .pointtopologyarray import PointTopologyArray
from .raggedcontiguousarray import RaggedContiguousArray
from .raggedindexedarray import RaggedIndexedArray
from .raggedindexedcontiguousarray import RaggedIndexedContiguousArray
from .sparsearray import SparseArray
from .subsampledarray import SubsampledArray

from .data import Data



import atexit
import datetime
import faulthandler
import os
import tempfile
import unittest

faulthandler.enable()  # to debug seg faults and timeouts

import cfdm

n_tmpfiles = 3
tmpfiles = [
    tempfile.mkstemp("_test_netCDF.nc", dir=os.getcwd())[1]
    for i in range(n_tmpfiles)
]
(tempfile1, tempfile2, tempfile3) = tmpfiles


def _remove_tmpfiles():
    """Remove temporary files created during tests."""
    for f in tmpfiles:
        try:
            os.remove(f)
        except OSError:
            pass


atexit.register(_remove_tmpfiles)


class NetCDFTest(unittest.TestCase):
    """Unit test for the NetCDF class."""

    def setUp(self):
        """Preparations called immediately before each test method."""
        # Disable log messages to silence expected warnings
        cfdm.log_level("DISABLE")
        # Note: to enable all messages for given methods, lines or
        # calls (those without a 'verbose' option to do the same)
        # e.g. to debug them, wrap them (for methods, start-to-end
        # internally) as follows:
        #
        # cfdm.log_level('DEBUG')
        # < ... test code ... >
        # cfdm.log_level('DISABLE')

        nc_group_structure_names = [
            None,
            "/",
            "group/...",
            "group/",
            "group/.../",
            "/group/.../",
        ]
        self.nc_grouped_dimension_names = [
            obj.replace("...", "ncdim")
            for obj in nc_group_structure_names
            if obj is not None
        ]
        self.nc_grouped_variable_names = [
            obj.replace("...", "ncvar")
            for obj in nc_group_structure_names
            if obj is not None
        ]

    def test_netCDF_variable_dimension(self):
        """Test variable and dimension access NetCDF methods."""
        f = cfdm.Field()

        f.nc_set_variable("qwerty")
        self.assertTrue(f.nc_has_variable())
        self.assertEqual(f.nc_get_variable(), "qwerty")
        self.assertEqual(f.nc_get_variable(default=None), "qwerty")
        self.assertEqual(f.nc_del_variable(), "qwerty")
        self.assertFalse(f.nc_has_variable())
        self.assertIsNone(f.nc_get_variable(default=None))
        self.assertIsNone(f.nc_del_variable(default=None))

        f.nc_set_variable("/ncvar")
        self.assertEqual(f.nc_get_variable(), "ncvar")

        f.nc_set_variable("/ncvar/qwerty")
        self.assertEqual(f.nc_get_variable(), "/ncvar/qwerty")

        for nc_var_name in self.nc_grouped_variable_names:
            with self.assertRaises(ValueError):
                f.nc_set_variable(nc_var_name)

        d = cfdm.DomainAxis()

        d.nc_set_dimension("qwerty")
        self.assertTrue(d.nc_has_dimension())
        self.assertEqual(d.nc_get_dimension(), "qwerty")
        self.assertEqual(d.nc_get_dimension(default=None), "qwerty")
        self.assertEqual(d.nc_del_dimension(), "qwerty")
        self.assertFalse(d.nc_has_dimension())
        self.assertIsNone(d.nc_get_dimension(default=None))
        self.assertIsNone(d.nc_del_dimension(default=None))

        d.nc_set_dimension("/ncdim")
        self.assertEqual(d.nc_get_dimension(), "ncdim")

        d.nc_set_dimension("/ncdim/qwerty")
        self.assertEqual(d.nc_get_dimension(), "/ncdim/qwerty")

        for nc_dim_name in self.nc_grouped_dimension_names:
            with self.assertRaises(ValueError):
                d.nc_set_dimension(nc_dim_name)

        d = cfdm.Count()

        d.nc_set_sample_dimension("qwerty")
        self.assertTrue(d.nc_has_sample_dimension())
        self.assertEqual(d.nc_get_sample_dimension(), "qwerty")
        self.assertEqual(d.nc_get_sample_dimension(default=None), "qwerty")
        self.assertEqual(d.nc_del_sample_dimension(), "qwerty")
        self.assertFalse(d.nc_has_sample_dimension())
        self.assertIsNone(d.nc_get_sample_dimension(default=None))
        self.assertIsNone(d.nc_del_sample_dimension(default=None))

        d.nc_set_sample_dimension("/ncdim")
        self.assertEqual(d.nc_get_sample_dimension(), "ncdim")

        d.nc_set_sample_dimension("/ncdim/qwerty")
        self.assertEqual(d.nc_get_sample_dimension(), "/ncdim/qwerty")

        for nc_dim_name in self.nc_grouped_dimension_names:
            with self.assertRaises(ValueError):
                d.nc_set_sample_dimension(nc_dim_name)

        # ------------------------------------------------------------
        # Global attributes
        # ------------------------------------------------------------
        # values keyword
        f = cfdm.Field()

        f.nc_set_global_attribute("Conventions", "CF-1.8")
        f.nc_set_global_attribute("project")
        f.nc_set_global_attribute("foo")
        f.set_property("Conventions", "Y")
        f.set_property("project", "X")
        self.assertEqual(
            f.nc_global_attributes(values=True),
            {"Conventions": "CF-1.8", "project": "X", "foo": None},
        )

        f = cfdm.Field()
        self.assertEqual(f.nc_clear_global_attributes(), {})

        f.nc_set_global_attribute("Conventions")
        f.nc_set_global_attribute("project", "X")
        self.assertEqual(
            f.nc_global_attributes(), {"Conventions": None, "project": "X"}
        )

        f.nc_set_global_attribute("project")
        f.nc_set_global_attribute("comment", None)
        self.assertEqual(
            f.nc_global_attributes(),
            {"Conventions": None, "project": None, "comment": None},
        )

        self.assertEqual(
            f.nc_clear_global_attributes(),
            {"Conventions": None, "project": None, "comment": None},
        )
        self.assertEqual(f.nc_global_attributes(), {})

        f.nc_set_global_attribute("Conventions")
        f.nc_set_global_attribute("project")
        self.assertEqual(
            f.nc_global_attributes(), {"Conventions": None, "project": None}
        )

        _ = f.nc_clear_global_attributes()
        f.nc_set_global_attributes({})
        self.assertEqual(f.nc_global_attributes(), {})

        f.nc_set_global_attributes({"comment": 123}, copy=False)
        self.assertEqual(f.nc_global_attributes(), {"comment": 123})

        f.nc_set_global_attributes({"comment": None, "foo": "bar"})
        self.assertEqual(
            f.nc_global_attributes(), {"comment": None, "foo": "bar"}
        )

        f = cfdm.Field()
        f.set_properties({"foo": "bar", "comment": "variable comment"})
        f.nc_set_variable("tas")
        d = f.set_construct(cfdm.DomainAxis(2))
        f.set_data(cfdm.Data([8, 9]), axes=[d])

        f2 = f.copy()
        f2.nc_set_variable("ua")

        cfdm.write(
            [f, f2],
            tempfile1,
            file_descriptors={"comment": "global comment", "qwerty": "asdf"},
        )

        g = cfdm.read(tempfile1)
        self.assertEqual(len(g), 2)

        for x in g:
            self.assertEqual(
                x.properties(),
                {
                    "comment": "variable comment",
                    "foo": "bar",
                    "qwerty": "asdf",
                    "Conventions": "CF-" + cfdm.CF(),
                },
            )
            self.assertEqual(
                x.nc_global_attributes(),
                {
                    "comment": "global comment",
                    "qwerty": None,
                    "Conventions": None,
                },
            )

        cfdm.write(g, tempfile2)
        h = cfdm.read(tempfile2)
        for x, y in zip(h, g):
            self.assertEqual(x.properties(), y.properties())
            self.assertEqual(
                x.nc_global_attributes(), y.nc_global_attributes()
            )
            self.assertTrue(x.equals(y, verbose=3))
            self.assertTrue(y.equals(x, verbose=3))

        g[1].nc_set_global_attribute("comment", "different comment")
        cfdm.write(g, tempfile3)
        h = cfdm.read(tempfile3)
        for x, y in zip(h, g):
            self.assertEqual(x.properties(), y.properties())
            self.assertEqual(
                x.nc_global_attributes(),
                {"comment": None, "qwerty": None, "Conventions": None},
            )
            self.assertTrue(x.equals(y, verbose=3))
            self.assertTrue(y.equals(x, verbose=3))

    def test_netCDF_geometry_variable(self):
        """Test geometry variable access and (un)setting methods."""
        f = cfdm.Field()

        f.nc_set_geometry_variable("qwerty")
        self.assertTrue(f.nc_has_geometry_variable())
        self.assertEqual(f.nc_get_geometry_variable(), "qwerty")
        self.assertEqual(f.nc_get_geometry_variable(default=None), "qwerty")
        self.assertEqual(f.nc_del_geometry_variable(), "qwerty")
        self.assertFalse(f.nc_has_geometry_variable())
        self.assertIsNone(f.nc_get_geometry_variable(default=None))
        self.assertIsNone(f.nc_del_geometry_variable(default=None))

        f.nc_set_geometry_variable("/ncvar")
        self.assertEqual(f.nc_get_geometry_variable(), "ncvar")

        f.nc_set_geometry_variable("/ncvar/qwerty")
        self.assertEqual(f.nc_get_geometry_variable(), "/ncvar/qwerty")

        for nc_var_name in self.nc_grouped_variable_names:
            with self.assertRaises(ValueError):
                f.nc_set_geometry_variable(nc_var_name)

    def test_netCDF_group_attributes(self):
        """Test group attributes access and (un)setting methods."""
        f = cfdm.Field()

        attrs = f.nc_group_attributes()
        self.assertIsInstance(attrs, dict)
        self.assertFalse(attrs)

        attrs = f.nc_clear_group_attributes()
        self.assertIsInstance(attrs, dict)
        self.assertFalse(attrs)

        attrs = f.nc_group_attributes()
        self.assertIsInstance(attrs, dict)
        self.assertFalse(attrs)

        f.nc_set_group_attributes({"comment": "somthing"})
        attrs = f.nc_group_attributes()
        self.assertEqual(attrs, {"comment": "somthing"})

        attrs = f.nc_clear_group_attributes()
        self.assertEqual(attrs, {"comment": "somthing"})

        attrs = f.nc_group_attributes()
        self.assertIsInstance(attrs, dict)
        self.assertFalse(attrs)

        f.nc_set_group_attributes({"comment": "something"})
        f.nc_set_group_attributes({"foo": "bar"})
        attrs = f.nc_group_attributes()
        self.assertEqual(attrs, {"comment": "something", "foo": "bar"})

        f.nc_clear_group_attributes()
        f.nc_set_group_attribute("foo", "bar")
        attrs = f.nc_group_attributes()
        self.assertEqual(attrs, {"foo": "bar"})
        f.nc_set_group_attribute("foo", "bar2")
        attrs = f.nc_group_attributes()
        self.assertEqual(attrs, {"foo": "bar2"})

        f.set_properties({"prop1": "value1", "comment": "variable comment"})
        f.nc_clear_group_attributes()
        f.nc_set_group_attributes({"comment": None, "foo": "bar"})
        self.assertEqual(
            f.nc_group_attributes(values=True),
            {"comment": "variable comment", "foo": "bar"},
        )

    def test_netCDF_dimension_groups(self):
        """Test dimension groups access and (un)setting methods."""
        d = cfdm.DomainAxis()

        d.nc_set_dimension("ncdim")

        with self.assertRaises(ValueError):
            d.nc_set_dimension_groups(["/forecast"])

        attrs = d.nc_dimension_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertFalse(attrs)

        attrs = d.nc_clear_dimension_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertFalse(attrs)

        attrs = d.nc_dimension_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertFalse(attrs)

        d.nc_set_dimension_groups(["forecast", "model"])
        attrs = d.nc_dimension_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertEqual(attrs, ("forecast", "model"))
        self.assertEqual(d.nc_get_dimension(), "/forecast/model/ncdim")

        attrs = d.nc_clear_dimension_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertEqual(attrs, ("forecast", "model"))

        attrs = d.nc_dimension_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertFalse(attrs)
        self.assertEqual(d.nc_get_dimension(), "ncdim")

        d.nc_set_dimension("ncdim")
        attrs = d.nc_dimension_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertEqual(attrs, ())

        d.nc_set_dimension("/ncdim")
        attrs = d.nc_dimension_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertFalse(attrs)

        d.nc_set_dimension("/forecast/model/ncdim")
        attrs = d.nc_dimension_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertEqual(attrs, ("forecast", "model"))

        d.nc_del_dimension()
        attrs = d.nc_dimension_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertFalse(attrs)

    def test_netCDF_variable_groups(self):
        """Test variable groups access and (un)setting methods."""
        f = cfdm.Field()

        f.nc_set_variable("ncdim")

        with self.assertRaises(ValueError):
            f.nc_set_variable_groups(["/forecast"])

        attrs = f.nc_variable_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertFalse(attrs)

        attrs = f.nc_clear_variable_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertFalse(attrs)

        attrs = f.nc_variable_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertFalse(attrs)

        f.nc_set_variable_groups(["forecast", "model"])
        attrs = f.nc_variable_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertEqual(attrs, ("forecast", "model"))
        self.assertEqual(f.nc_get_variable(), "/forecast/model/ncdim")

        attrs = f.nc_clear_variable_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertEqual(attrs, ("forecast", "model"))

        attrs = f.nc_variable_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertFalse(attrs)
        self.assertEqual(f.nc_get_variable(), "ncdim")

        f.nc_set_variable("ncdim")
        attrs = f.nc_variable_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertEqual(attrs, ())

        f.nc_set_variable("/ncdim")
        attrs = f.nc_variable_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertFalse(attrs)

        f.nc_set_variable("/forecast/model/ncdim")
        attrs = f.nc_variable_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertEqual(attrs, ("forecast", "model"))

        f.nc_del_variable()
        attrs = f.nc_variable_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertFalse(attrs)

        with self.assertRaises(ValueError):
            f.nc_set_variable_groups(["forecast", "model"])

    def test_netCDF_geometry_variable_groups(self):
        """Test geometry variable groups access NetCDF methods."""
        f = cfdm.Field()

        f.nc_set_geometry_variable("ncvar")

        with self.assertRaises(ValueError):
            f.nc_set_geometry_variable_groups(["/forecast"])

        attrs = f.nc_geometry_variable_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertFalse(attrs)

        attrs = f.nc_clear_geometry_variable_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertFalse(attrs)

        attrs = f.nc_geometry_variable_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertFalse(attrs)

        f.nc_set_geometry_variable_groups(["forecast", "model"])
        attrs = f.nc_geometry_variable_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertEqual(attrs, ("forecast", "model"))
        self.assertEqual(f.nc_get_geometry_variable(), "/forecast/model/ncvar")

        attrs = f.nc_clear_geometry_variable_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertEqual(attrs, ("forecast", "model"))

        attrs = f.nc_geometry_variable_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertFalse(attrs)
        self.assertEqual(f.nc_get_geometry_variable(), "ncvar")

        f.nc_set_geometry_variable("ncvar")
        attrs = f.nc_geometry_variable_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertEqual(attrs, ())

        f.nc_set_geometry_variable("/ncvar")
        attrs = f.nc_geometry_variable_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertFalse(attrs)

        f.nc_set_geometry_variable("/forecast/model/ncvar")
        attrs = f.nc_geometry_variable_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertEqual(attrs, ("forecast", "model"))

        f.nc_del_geometry_variable()
        attrs = f.nc_geometry_variable_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertFalse(attrs)

        with self.assertRaises(ValueError):
            f.nc_set_geometry_variable_groups(["forecast", "model"])

    def test_netCDF_sample_dimension_groups(self):
        """Test sample dimension groups access NetCDF methods."""
        c = cfdm.Count()

        c.nc_set_sample_dimension("ncvar")

        with self.assertRaises(ValueError):
            c.nc_set_sample_dimension_groups(["/forecast"])

        attrs = c.nc_sample_dimension_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertFalse(attrs)

        attrs = c.nc_clear_sample_dimension_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertFalse(attrs)

        attrs = c.nc_sample_dimension_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertFalse(attrs)

        c.nc_set_sample_dimension_groups(["forecast", "model"])
        attrs = c.nc_sample_dimension_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertEqual(attrs, ("forecast", "model"))
        self.assertEqual(c.nc_get_sample_dimension(), "/forecast/model/ncvar")

        attrs = c.nc_clear_sample_dimension_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertEqual(attrs, ("forecast", "model"))

        attrs = c.nc_sample_dimension_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertFalse(attrs)
        self.assertEqual(c.nc_get_sample_dimension(), "ncvar")

        c.nc_set_sample_dimension("ncvar")
        attrs = c.nc_sample_dimension_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertEqual(attrs, ())

        c.nc_set_sample_dimension("/ncvar")
        attrs = c.nc_sample_dimension_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertFalse(attrs)

        c.nc_set_sample_dimension("/forecast/model/ncvar")
        attrs = c.nc_sample_dimension_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertEqual(attrs, ("forecast", "model"))

        c.nc_del_sample_dimension()
        attrs = c.nc_sample_dimension_groups()
        self.assertIsInstance(attrs, tuple)
        self.assertFalse(attrs)

        with self.assertRaises(ValueError):
            c.nc_set_sample_dimension_groups(["forecast", "model"])

    def test_netCDF_field_components(self):
        """Test field component access and (un)setting methods."""
        # Geometries
        f = cfdm.example_field(6)

        for component in ("interior_ring", "node_count", "part_node_count"):
            f.nc_set_component_variable(component, "ncvar")
            f.nc_set_component_variable_groups(component, ["forecast"])

            f.nc_clear_component_variable_groups(component)
            f.nc_del_component_variable(component)

            f.nc_del_component_variable(component)
            f.nc_clear_component_variable_groups(component)

            f.nc_set_component_variable(component, "ncvar")
            f.nc_set_component_variable_groups(component, ["forecast"])

        for component in ("interior_ring", "part_node_count"):
            f.nc_set_component_dimension(component, "ncvar")
            f.nc_set_component_dimension_groups(component, ["forecast"])

            f.nc_clear_component_dimension_groups(component)
            f.nc_del_component_dimension(component)

            f.nc_del_component_dimension(component)
            f.nc_clear_component_dimension_groups(component)

            f.nc_set_component_dimension(component, "ncvar")
            f.nc_set_component_dimension_groups(component, ["forecast"])

        # Compression: indexed and contiguous
        f = cfdm.example_field(4)
        f.compress("indexed_contiguous", inplace=True)

        for component in ("count", "index"):
            f.nc_set_component_variable(component, "ncvar")
            f.nc_set_component_variable_groups(component, ["forecast"])

            f.nc_clear_component_variable_groups(component)
            f.nc_del_component_variable(component)

            f.nc_del_component_variable(component)
            f.nc_clear_component_variable_groups(component)

            f.nc_set_component_variable(component, "ncvar")
            f.nc_set_component_variable_groups(component, ["forecast"])

        for component in ("count", "index"):
            f.nc_set_component_dimension(component, "ncvar")
            f.nc_set_component_dimension_groups(component, ["forecast"])

            f.nc_clear_component_dimension_groups(component)
            f.nc_del_component_dimension(component)

            f.nc_del_component_dimension(component)
            f.nc_clear_component_dimension_groups(component)

            f.nc_set_component_dimension(component, "ncvar")
            f.nc_set_component_dimension_groups(component, ["forecast"])

        for component in ("count", "index"):
            f.nc_set_component_sample_dimension(component, "ncvar")
            f.nc_set_component_sample_dimension_groups(component, ["forecast"])

            f.nc_clear_component_sample_dimension_groups(component)
            f.nc_del_component_sample_dimension(component)

            f.nc_del_component_sample_dimension(component)
            f.nc_clear_component_sample_dimension_groups(component)

            f.nc_set_component_sample_dimension(component, "ncvar")
            f.nc_set_component_sample_dimension_groups(component, ["forecast"])

        # Compression: gathered
        component = "list"

        # Expected exceptions
        for component in ("list", "node_count"):
            with self.assertRaises(ValueError):
                f.nc_set_component_dimension(component, "ncvar")

            with self.assertRaises(ValueError):
                f.nc_del_component_dimension(component)

            with self.assertRaises(ValueError):
                f.nc_set_component_dimension_groups(component, "ncvar")

            with self.assertRaises(ValueError):
                f.nc_clear_component_dimension_groups(component)

            with self.assertRaises(ValueError):
                f.nc_set_component_sample_dimension(component, "ncvar")

            with self.assertRaises(ValueError):
                f.nc_del_component_sample_dimension(component)

            with self.assertRaises(ValueError):
                f.nc_set_component_sample_dimension_groups(component, "ncvar")

            with self.assertRaises(ValueError):
                f.nc_clear_component_sample_dimension_groups(component)

        # Expected exceptions
        for component in ("WRONG",):
            with self.assertRaises(ValueError):
                f.nc_set_component_variable(component, "ncvar")

            with self.assertRaises(ValueError):
                f.nc_del_component_variable(component)

            with self.assertRaises(ValueError):
                f.nc_set_component_variable_groups(component, "ncvar")

            with self.assertRaises(ValueError):
                f.nc_clear_component_variable_groups(component)


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cfdm.environment()
    print("")
    unittest.main(verbosity=2)



from copy import deepcopy
from urllib.parse import urlparse

from s3fs import S3FileSystem

from ...functions import abspath


class DeprecationError(Exception):
    """Deprecation error."""

    pass


class FileArrayMixin:
    """Mixin class for a file container of an array.

    .. versionadded:: (cfdm) 1.10.1.0

    """

    def __repr__(self):
        """Called by the `repr` built-in function.

        x.__repr__() <==> repr(x)

        """
        return f"<{self.__class__.__name__}{self.shape}: {self}>"

    def __str__(self):
        """Called by the `str` built-in function.

        x.__str__() <==> str(x)

        """
        filenames = self.get_filenames()
        addresses = self.get_addresses()
        if len(filenames) == 1:
            filenames = filenames[0]
            addresses = addresses[0]

        return f"{filenames}, {addresses}"

    def __dask_tokenize__(self):
        """Return a value fully representative of the object.

        .. versionadded:: (cfdm) 1.11.2.0

        """
        return (
            self.__class__,
            self.shape,
            self.get_filenames(),
            self.get_addresses(),
        )

    @property
    def dtype(self):
        """Data-type of the array."""
        return self._get_component("dtype")

    @property
    def shape(self):
        """Shape of the array."""
        return self._get_component("shape")

    def close(self, dataset):
        """Close the dataset containing the data."""
        raise NotImplementedError(
            f"Must implement {self.__class__.__name__}.close"
        )  # pragma: no cover

    def get_address(self, default=AttributeError()):
        """The name of the file containing the array.

        If there are multiple files then an `AttributeError` is
        raised by default.

        .. versionadded:: (cfdm) 1.10.1.0

        :Parameters:

            default: optional
                Return the value of the *default* parameter if there
                is no file.

                {{default Exception}}

        :Returns:

            `str`
                The file name.

        """
        addresses = self.get_addresses()
        n = len(addresses)
        if n == 1:
            return addresses[0]

        if default is None:
            return

        return self._default(
            default, f"{self.__class__.__name__} has no unique file address"
        )

    def get_addresses(self):
        """Return the names of the data addresses in the files.

        If there are multiple addresses then they correspond, in
        order, to the files returned by `get_filenames`

        .. versionadded:: (cfdm) 1.10.0.2

        .. seealso:: `get_filenames`

        :Returns:

            `tuple`
                The addresses.

        """
        return self._get_component("address", ())

    def get_filename(self, default=AttributeError()):
        """The name of the file containing the array.

        If there are multiple files then an `AttributeError` is
        raised by default.

        .. versionadded:: (cfdm) 1.10.0.2

        :Parameters:

            default: optional
                Return the value of the *default* parameter if there
                is no file or there is more than one file.

                {{default Exception}}

        :Returns:

            `str`
                The file name.

        """
        filenames = self._get_component("filename", ())
        if len(filenames) == 1:
            return filenames[0]

        if default is None:
            return

        return self._default(
            default, f"{self.__class__.__name__} has no files"
        )

    def get_filenames(self):
        """Return the names of files containing the data.

        If multiple files are returned then it is assumed that any
        one of them may contain the data, and when the data are
        requested an attempt to open file is made, in order, and the
        data is read from the first success.

        .. versionadded:: (cfdm) 1.10.0.2

        .. seealso:: `get_addresses`

        :Returns:

            `tuple`
                The filenames, in absolute form.

        """
        return tuple([abspath(f) for f in self._get_component("filename", ())])

    def get_format(self):
        """The format of the files.

        .. versionadded:: (cfdm) 1.10.1.0

        .. seealso:: `get_address`, `get_filename`, `get_formats`

        :Returns:

            `str`
                 The file format.

        """
        raise NotImplementedError(
            f"Must implement {self.__class__.__name__}.get_format"
        )  # pragma: no cover

    def get_formats(self):
        """Return the format of the files.

        .. versionadded:: (cfdm) 1.10.1.0

        .. seealso:: `get_format`, `get_filenames`, `get_addresses`

        :Returns:

            `tuple`
                The fragment file formats.

        """
        return (self.get_format(),) * len(self.get_filenames())

    def get_missing_values(self):
        """The missing values of the data.

        Deprecated at version 1.11.2.0. Use `get_attributes` instead.

        """
        raise DeprecationError(
            f"{self.__class__.__name__}.get_missing_values was deprecated "
            "at version 1.11.2.0 and is no longer available. "
            f"Use {self.__class__.__name__}.get_attributes instead."
        )  # pragma: no cover

    def get_storage_options(
        self, create_endpoint_url=True, filename=None, parsed_filename=None
    ):
        """Return `s3fs.S3FileSystem` options for accessing S3 files.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameters:

            create_endpoint_url: `bool`, optional
                If True, the default, then create an
                ``'endpoint_url'`` option if and only if one has not
                already been provided. See *filename* and
                *parsed_filename* for details.

            filename: `str`, optional
                Used to set the ``'endpoint_url'`` option if it has
                not been previously defined. Ignored if
                *parsed_filename* has been set.

            parsed_filename: `urllib.parse.ParseResult`, optional
                Used to set the ``'endpoint_url'`` option if it has
                not been previously defined. By default the
                ``'endpoint_url'`` option, if required, is set from
                the file name returned by `get_filename`.

        :Returns:

            `dict` or `None`
                The `s3fs.S3FileSystem` options.

        **Examples**

        >>> f.get_filename()
        's3://store/data/file.nc'
        >>> f.get_storage_options(create_endpoint_url=False)
        {}
        >>> f.get_storage_options()
        {'endpoint_url': 'https://store'}
        >>> f.get_storage_options(filename='s3://other-store/data/file.nc')
        {'endpoint_url': 'https://other-store'}
        >>> f.get_storage_options(create_endpoint_url=False,
        ...                       filename='s3://other-store/data/file.nc')
        {}

        >>> f.get_storage_options()
        {'key': 'scaleway-api-key...',
         'secret': 'scaleway-secretkey...',
         'endpoint_url': 'https://s3.fr-par.scw.cloud',
         'client_kwargs': {'region_name': 'fr-par'}}

        """
        storage_options = self._get_component("storage_options", None)
        if not storage_options:
            storage_options = {}
        else:
            storage_options = deepcopy(storage_options)

        client_kwargs = storage_options.get("client_kwargs", {})
        if (
            create_endpoint_url
            and "endpoint_url" not in storage_options
            and "endpoint_url" not in client_kwargs
        ):
            if parsed_filename is None:
                if filename is None:
                    try:
                        filename = self.get_filename()
                    except AttributeError:
                        pass
                    else:
                        parsed_filename = urlparse(filename)
                else:
                    parsed_filename = urlparse(filename)

            if parsed_filename is not None and parsed_filename.scheme == "s3":
                # Derive endpoint_url from filename
                storage_options["endpoint_url"] = (
                    f"https://{parsed_filename.netloc}"
                )

        return storage_options

    def open(self, func, *args, **kwargs):
        """Return a dataset file object and address.

        When multiple files have been provided an attempt is made to
        open each one, in the order stored, and a file object is
        returned from the first file that exists.

        .. versionadded:: (cfdm) 1.10.1.0

        :Parameters:

            func: callable
                Function that opens a file.

            args, kwargs: optional
                Optional arguments to *func*.

        :Returns:

            2-`tuple`
                The file object for the dataset, and the address of
                the data within the file.

        """
        # Loop round the files, returning as soon as we find one that
        # works.
        filenames = self.get_filenames()
        for filename, address in zip(filenames, self.get_addresses()):
            url = urlparse(filename)
            if url.scheme == "file":
                # Convert a file URI into an absolute path
                filename = url.path
            elif url.scheme == "s3":
                # Create an openable S3 file object
                storage_options = self.get_storage_options(
                    create_endpoint_url=True, parsed_filename=url
                )
                fs = S3FileSystem(**storage_options)
                filename = fs.open(url.path[1:], "rb")

            try:
                dataset = func(filename, *args, **kwargs)
            except FileNotFoundError:
                continue
            except RuntimeError as error:
                raise RuntimeError(f"{error}: {filename}")

            # Successfully opened a dataset, so return.
            return dataset, address

        if len(filenames) == 1:
            raise FileNotFoundError(f"No such file: {filenames[0]}")

        raise FileNotFoundError(f"No such files: {filenames}")



from ..fullarray import FullArray
from .mixin import FragmentArrayMixin


class FragmentUniqueValueArray(FragmentArrayMixin, FullArray):
    """A fragment of aggregated data that has a single unique value.

    .. versionadded:: (cfdm) 1.12.0.0

    """

    def __init__(
        self,
        unique_value=None,
        dtype=None,
        shape=None,
        unpack_aggregated_data=True,
        aggregated_attributes=None,
        attributes=None,
        source=None,
        copy=True,
    ):
        """**Initialisation**

        :Parameters:

            unique_value: scalar
                The unique value for the fragment.

            dtype: `numpy.dtype`
                The data type of the aggregated array. May be `None`
                if the numpy data-type is not known (which can be the
                case for netCDF string types, for example). This may
                differ from the data type of the netCDF fragment
                variable.

            shape: `tuple`
                The shape of the fragment within the aggregated
                array. This may differ from the shape of the netCDF
                fragment variable in that the latter may have fewer
                size 1 dimensions.

            {{init attributes: `dict` or `None`, optional}}

            {{aggregated_units: `str` or `None`, optional}}

            {{aggregated_calendar: `str` or `None`, optional}}

            {{init source: optional}}

            {{init copy: `bool`, optional}}

        """
        super().__init__(
            fill_value=unique_value,
            dtype=dtype,
            shape=shape,
            attributes=None,
            source=source,
            copy=False,
        )

        if source is not None:

            try:
                aggregated_attributes = source._get_component(
                    "aggregated_attributes", None
                )
            except AttributeError:
                aggregated_attributes = None

            try:
                unpack_aggregated_data = source._get_component(
                    "unpack_aggregated_data", True
                )
            except AttributeError:
                unpack_aggregated_data = True

        self._set_component(
            "unpack_aggregated_data",
            unpack_aggregated_data,
            copy=False,
        )
        if aggregated_attributes is not None:
            self._set_component(
                "aggregated_attributes", aggregated_attributes, copy=copy
            )



import atexit
import datetime
import faulthandler
import os
import tempfile
import unittest

import numpy as np

faulthandler.enable()  # to debug seg faults and timeouts

import cfdm

n_tmpfiles = 1
tmpfiles = [
    tempfile.mkstemp("_test_string.nc", dir=os.getcwd())[1]
    for i in range(n_tmpfiles)
]
(tempfile,) = tmpfiles


def _remove_tmpfiles():
    """Remove temporary files created during tests."""
    for f in tmpfiles:
        try:
            os.remove(f)
        except OSError:
            pass


atexit.register(_remove_tmpfiles)


class StringTest(unittest.TestCase):
    """Test constructs with underlying arrays of string data type."""

    def setUp(self):
        """Preparations called immediately before each test method."""
        # Disable log messages to silence expected warnings
        cfdm.log_level("DISABLE")
        # Note: to enable all messages for given methods, lines or
        # calls (those without a 'verbose' option to do the same)
        # e.g. to debug them, wrap them (for methods, start-to-end
        # internally) as follows:
        #
        # cfdm.log_level('DEBUG')
        # < ... test code ... >
        # cfdm.log_level('DISABLE')

    def test_STRING(self):
        """Test constructs with underlying string type arrays."""
        for array in (
            np.ma.array(list("abcdefghij"), dtype="S"),
            np.ma.array(
                ["a", "b1", "c12", "d123", "e1234", "f", "g", "h", "i", "j"],
                dtype="S",
            ),
        ):
            # Initialize the field
            tas = cfdm.Field(
                properties={
                    "project": "research",
                    "standard_name": "air_temperature",
                    "units": "K",
                }
            )

            # Create and set domain axes
            tas.set_construct(cfdm.DomainAxis(1))
            tas.set_construct(cfdm.DomainAxis(1))
            axis_Y = tas.set_construct(cfdm.DomainAxis(10))
            axis_X = tas.set_construct(cfdm.DomainAxis(9))

            # Set the field data
            tas.set_data(
                cfdm.Data(np.arange(90.0).reshape(10, 9)),
                axes=[axis_Y, axis_X],
            )

            # Create and set the dimension coordinates
            dimension_coordinate_Y = cfdm.DimensionCoordinate(
                properties={
                    "standard_name": "grid_latitude",
                    "units": "degrees",
                },
                data=cfdm.Data(np.arange(10.0)),
                bounds=cfdm.Bounds(
                    data=cfdm.Data(np.arange(20).reshape(10, 2))
                ),
            )

            dimension_coordinate_X = cfdm.DimensionCoordinate(
                properties={
                    "standard_name": "grid_longitude",
                    "units": "degrees",
                },
                data=cfdm.Data(np.arange(9.0)),
                bounds=cfdm.Bounds(
                    data=cfdm.Data(np.arange(18).reshape(9, 2))
                ),
            )

            tas.set_construct(dimension_coordinate_Y, axes=[axis_Y])
            tas.set_construct(dimension_coordinate_X, axes=[axis_X])

            # Create and set the auxiliary coordinates
            array[0] = np.ma.masked
            aux0 = cfdm.AuxiliaryCoordinate(
                properties={"long_name": "Grid latitude name"},
                data=cfdm.Data(array),
            )

            tas.set_construct(aux0, axes=[axis_Y])

            cfdm.write(tas, tempfile)

            tas1 = cfdm.read(tempfile)[0]

            aux1 = tas1.constructs.filter_by_identity(
                "long_name=Grid latitude name"
            ).value()
            self.assertEqual(aux0.data.shape, array.shape, aux0.data.shape)
            self.assertEqual(aux1.data.shape, array.shape, aux1.data.shape)


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cfdm.environment()
    print("")
    unittest.main(verbosity=2)



from itertools import accumulate, product

import numpy as np

from .abstract import RaggedArray
from .mixin import CompressedArrayMixin


class RaggedIndexedContiguousArray(CompressedArrayMixin, RaggedArray):
    """An underlying indexed contiguous ragged array.

    A collection of features, each of which is sequence of (vertical)
    profiles, stored using an indexed contiguous ragged array combines
    all feature elements along a single dimension (the "sample
    dimension") such that a contiguous ragged array representation is
    used for each profile and the indexed ragged array representation
    to organise the profiles into timeseries.

    The information needed to uncompress the data is stored in a
    "count variable" that gives the size of each profile; and in a
    "index variable" that specifies the feature that each profile
    belongs to.

    It is assumed that the compressed dimensions are the two left-most
    dimensions in the compressed array.

    See CF section 9 "Discrete Sampling Geometries".

    .. versionadded:: (cfdm) 1.7.0

    """

    def __init__(
        self,
        compressed_array=None,
        shape=None,
        size=None,
        ndim=None,
        count_variable=None,
        index_variable=None,
        source=None,
        copy=True,
    ):
        """**Initialisation**

        :Parameters:

            compressed_array: array_like
                The compressed array.

            shape: `tuple`
                The shape of the uncompressed array.

            count_variable: `Count`
                The count variable required to uncompress the data,
                corresponding to a CF-netCDF count variable.

            index_variable: `Index`
                The index variable required to uncompress the data,
                corresponding to a CF-netCDF CF-netCDF index variable.

            {{init source: optional}}

                .. versionadded:: (cfdm) 1.10.0.0

            {{init copy: `bool`, optional}}

                .. versionadded:: (cfdm) 1.10.0.0

            size: `int`
                Deprecated at version 1.10.0.0. Ignored if set.

                Number of elements in the uncompressed array.

            ndim: `int`
                Deprecated at version 1.10.0.0. Ignored if set.

                The number of uncompressed array dimensions.

        """
        super().__init__(
            compressed_array=compressed_array,
            shape=shape,
            count_variable=count_variable,
            index_variable=index_variable,
            compressed_dimensions={0: (0, 1, 2)},
            source=source,
            copy=copy,
        )

    def subarrays(self, shapes=-1):
        """Return descriptors for every subarray.

        Theses descriptors are used during subarray decompression.

        .. versionadded:: (cfdm) 1.10.0.0

        :Parameters:

            {{subarrays chunks: ``-1`` or sequence, optional}}

        :Returns:

             4-`tuple` of iterators
                Each iterable iterates over a particular descriptor
                from each subarray.

                1. The indices of the uncompressed array that
                   correspond to each subarray.

                2. The shape of each uncompressed subarray.

                3. The indices of the compressed array that correspond
                   to each subarray.

                4. The location of each subarray on the uncompressed
                   dimensions.

        **Examples**

        An original 3-d array with shape (2, 3, 4) comprising 2
        timeSeriesProfile features has been compressed as an indexed
        contiguous ragged array. The first feature has 3 profiles with
        counts of 2, 4, and 3 elements, at compressed locations (4,
        5), (0, 1, 2, 3), and (9, 10, 11) respectively. The second
        feature has 1 profile with a count of 3 elements, at
        compressed locations (6, 7, 8).

        >>> u_indices, u_shapes, c_indices, locations = x.subarrays()
        >>> for i in u_indices:
        ...    print(i)
        ...
        (slice(0, 1, None), slice(0, 1, None), slice(0, 4, None))
        (slice(0, 1, None), slice(1, 2, None), slice(0, 4, None))
        (slice(0, 1, None), slice(2, 3, None), slice(0, 4, None))
        (slice(1, 2, None), slice(0, 1, None), slice(0, 4, None))
        (slice(1, 2, None), slice(1, 2, None), slice(0, 4, None))
        (slice(1, 2, None), slice(2, 3, None), slice(0, 4, None))
        >>> for i in u_shapes
        ...    print(i)
        ...
        (1, 1, 4)
        (1, 1, 4)
        (1, 1, 4)
        (1, 1, 4)
        (1, 1, 4)
        (1, 1, 4)
        >>> for i in c_indices:
        ...    print(i)
        ...
        (slice(4, 6, None),)
        (slice(0, 4, None),)
        (slice(9, 12, None),)
        (slice(6, 9, None),)
        (slice(0, 0, None),)
        (slice(0, 0, None),)
        >>> for i in locations:
        ...    print(i)
        ...
        (0, 0, 0)
        (0, 1, 0)
        (0, 2, 0)
        (1, 0, 0)
        (1, 1, 0)
        (1, 2, 0)

        """
        d1, u_dims = self.compressed_dimensions().popitem()

        shapes = self.subarray_shapes(shapes)

        # The indices of the uncompressed array that correspond to
        # each subarray, the shape of each uncompressed subarray, and
        # the location of each subarray
        locations, u_shapes, u_indices = self._uncompressed_descriptors(
            u_dims, shapes
        )

        # The indices of the compressed array that correspond to each
        # subarray
        c_indices = []
        for d, size in enumerate(self.source().shape):
            if d == d1:
                index = np.array(self.get_index())
                unique = np.unique(index).tolist()
                count_partial_sums = np.cumsum(
                    np.array(self.get_count())
                ).tolist()

                max_n_profiles = self.shape[u_dims[1]]

                ind = []
                for i in unique:
                    # find the locations in the count array for the profiles
                    # in this feature.
                    profile_locations = np.where(index == i)[0]

                    for j in profile_locations:
                        if not j:
                            start = 0
                        else:
                            start = count_partial_sums[j - 1]

                        ind.append(slice(start, count_partial_sums[j]))

                    # Add zero-sized slices for this feature's "missing"
                    # profiles
                    ind.extend(
                        (slice(0, 0),)
                        * (max_n_profiles - profile_locations.size)
                    )

                c_indices.append(ind)
            else:
                if d < d1:
                    c = shapes[d]
                else:
                    c = shapes[d + 1]

                c = tuple(accumulate((0,) + c))
                c_indices.append([slice(i, j) for i, j in zip(c[:-1], c[1:])])

        return (
            product(*u_indices),
            product(*u_shapes),
            product(*c_indices),
            product(*locations),
        )
